/*********************************************************/
/** Name: seqcomp                                      **/
/** Version: 0.8                                       **/
/** Date: 2008-05-01								   **/	
/** Author: Laurent Lesnard							   **/
/** Depends: Sas, IML, and  %quotelst                  **/	
/** Description: Implementation of the dynamic hamming **/
/** distance proposed by the author to compare         **/
/** sequences in social sciences                       **/
/** Syntax: %seqcomp(data,varlist,export=0,		       **/
/**	out=dhamdist,outfile="c:\temp\dhamdist.txt",print=1**/
/**  -data: dataset to analyze                         **/
/**  -varlist: list of the variables to be analyzed    **/
/**            x1-x100 and x1--x100 are also accepted. **/
/**            States must be coded from 1 to K,       **/
/**            K being the number of states	           **/  
/**  -weight   name of the variable containting weights**/  
/**            0 by default means no weights are used  **/
/**  -export: if export=0 (default) then the           **/ 
/**           dissimilarity matrix is saved as a       **/
/**           sas dataset whose name is set by default **/
/**           to "dhamdist" but can be changed through **/
/**           the out option. If export=1, the         **/
/**           dissimilarity matrix is exported into a  **/
/**           text file which can be imported into     **/
/**           ClustanGraphics for instance             **/
/**  -out: optional, name of the dataset where results **/
/**        will be stored if export=0                  **/
/**  -outfile: file path to store the distance list    **/
/**            when export=1
/**  -print: by default transition matrices are printed**/
/**          in the output screen but a quiet mode     **/
/**          (print=0) is available                    **/
/** Note: two warning messages I'm working on to       **/
/**      suppress.                                     **/
/********************************************************/
/** Typical use:                                       **/
/********************************************************/
/**   in a sas workflow:                               **/  
%seqcomp(data.seq,episode1-episode100);
proc cluster data=dhamdist method=flexible beta=-0.3 print=20 outtree=treedham;
 id id;/*id is the name of the variable that identifies individuals*/
run;
/**   in a ClustanGraphics-like workflow                    **/
%seqcomp(data.seq,episode1-episode100,export=1,outfile="d:\data\dhamdist.txt");
/********************************************************/
/********************************************************/


/** Two macros to compile **/
%macro quotelst(str,quote=%str(%"),delim=%str( ));
	%local i quotelst;
	%let i=1;
	%do %while(%length(%scan(&str,&i,%str( ))) GT 0);
  		%if %length(&quotelst) EQ 0 %then %let quotelst=&quote.%scan(&str,&i,%str( ))&quote;
  		%else %let quotelst=&quotelst.&quote.%scan(&str,&i,%str( ))&quote;
  		%let i=%eval(&i + 1);
  		%if %length(%scan(&str,&i,%str( ))) GT 0 %then %let quotelst=&quotelst.&delim;
	%end;
	%unquote(&quotelst)
%mend quotelst;


%macro seqcomp(data,varlist,weight=0,export=0,out=dhamdist,outfile="c:\temp\dhamdist.txt",print=1);
options nonotes;
proc contents data=&data out=_des noprint;
run;
data _null_;
 set _des;
 call symput('N',put(nobs,6.));
run;
data _z;
 set &data;
 _wt=1;
run;
proc datasets library=work  nolist;
	delete _des;
quit;
/*submodule expand varlist*/
%let tiret1=%str( -);
%do %while (%index(&varlist,&tiret1)>0);
	%let long=%index(&varlist,&tiret1);
	%let varlist=%qsubstr(&varlist,1,&long-1)%str(-)%qsubstr(&varlist,&long+2);
%end;

%let tiret2=%str(- );
%do %while (%index(&varlist,&tiret2)>0);
	%let long=%index(&varlist,&tiret2);
	%let varlist=%qsubstr(&varlist,1,&long-1)%str(-)%qsubstr(&varlist,&long+2);
%end;

%if %index(&varlist,--)>0 %then %do;
	proc contents data=&data noprint; run;
%end;
%do %while (%index(&varlist,--)>0);
	%let i=1;
	%do %while(%index(%scan(&varlist,&i,%str( )),--)=0);
		%let i=%eval(&i+1);
	%end;
	%let long2=%length(%scan(&varlist,&i,%str( )));
	%let Vd=%scan(&varlist,&i);
	%let long=%index(&varlist,&vd);
	%let Vf=%scan(&varlist,&i+1);
	%local z; %local zz;
	%let z=%sysfunc(open(&data));
		%let posd=%sysfunc(varnum(&z,&vd));
		%let posf=%sysfunc(varnum(&z,&vf));
		%let j=%eval(&posd+1);
		%let blanc=%str( );
		%let ttevar=&vd&blanc;
		%do %while (&j<&posf);
			%let vj=%sysfunc(varname(&z,&j));
			%let ttevar=&ttevar&vj&blanc;
			%let j=%eval(&j+1);
		%end;
		%let ttevar=&ttevar&vf;
	%let zz=%sysfunc(close(&z));
	%let varlist=%qsubstr(&varlist,1,&long-1)%str(&ttevar)%qsubstr(&varlist,&long+&long2);
%end;

%do %while (%index(&varlist,-)>0);
	%let i=1;
	%do %while(%index(%scan(&varlist,&i,%str( )),-)=0);
		%let i=%eval(&i+1);
	%end;
	%let long2=%length(%scan(&varlist,&i,%str( )));
	%let Vd=%scan(&varlist,&i);
	%let long=%index(&varlist,&vd);
	%let Vf=%scan(&varlist,&i+1);
	%let k=%length(&Vd);
	%do %while(%verify(%substr(&vd,&k),0123456789)=0);
		%let posd=%substr(&vd,&k);
		%let k=%eval(&k-1);
	%end;
	%let k=%length(&Vf);
	%do %while(%verify(%substr(&vf,&k),0123456789)=0);
		%let posf=%substr(&vf,&k);
		%let k=%eval(&k-1);
	%end;
		%let j=%eval(&posd+1);
		%let blanc=%str( );
		%let ttevar=&vd&blanc;
		%do %while (&j<&posf);
			%let vj=%substr(&vd,1,&k);
			%let ttevar=&ttevar&vj&j&blanc;
			%let j=%eval(&j+1);
		%end;
		%let ttevar=&ttevar&vf;
	%let varlist=%qsubstr(&varlist,1,&long-1)%str(&ttevar)%qsubstr(&varlist,&long+&long2);
%end;
/*End sub-module*/
%local i; 
%let seqleng=0; 
%do %while(%scan(&varlist,&seqleng+1,%str( )) ne %str( ));
	%let seqleng = %eval(&seqleng+1);
%end; 
%put Number of sequences = &N;
%put Length of sequences = &seqleng;
%put Episode names = &varlist;
%if &weight=0 %then %put No weights are used;
%else %put Weights = &weight;
data _null_;
 set &data;
 retain maxstates maxx 1;
 maxstates=max(%quotelst(&varlist,quote=%str(),delim=%str(,)));
 if (_N_ > 1 ) then maxx=max(maxstates,lag(maxx));
 if _n_=&N then call symput('nstate',put(maxx,6.));
run;
%put Number of states = &nstate;
%let nlist=%eval(&n*(&n-1)/2);
/*%put &nlist;*/
proc iml SYMSIZE=5000;
 start subs(table);
 /*n=nrow(table);
 nbseq=ncol(table);*/
/*Calcul de nbseq matrice de transition*/
  subs&seqleng=repeat(0,&nstate,&nstate);
  %do k=1 %to %eval(&seqleng-1);
  	/*%put &k;*/
  	%if &k>1 %then %do;
		mark1old=mark1;
		mark2old=mark2;
	%end;
   		subs&k=repeat(0,&nstate,&nstate);
		MARK1=REPEAT(0,&nstate,&nstate);
		MARK2=REPEAT(0,&nstate,1);
 	do l=1 to &n; /*boucle sur les individus*/
/*Calcul des cots de substitutions*/
	   	i=table[l,&k];j=table[l,%eval(&k+1)];
        mark1[i,j]=sum(mark1[i,j],table[l,&seqleng+1]);
        mark2[i]=sum(mark2[i],table[l,&seqleng+1]);
 	end;
	do i=1 to &nstate;
        do j=1 to &nstate;
			%if &k=1 %then %do;subs&k[i,j]=4-2*max(0,sum(mark1[i,j]/mark2[i],mark1[j,i]/mark2[j])); %end;
			%else %do; subs&k[i,j]=4-max(0,sum(mark1[i,j]/mark2[i], mark1[j,i]/mark2[j], mark1old[i,j]/mark2old[i], mark1old[j,i]/mark2old[j])) ;%end;
			%if &k=%eval(&seqleng-1) %then %do;
				subs&seqleng[i,j]=4-2*max(0,sum(mark1old[i,j]/mark2old[i],mark1old[j,i]/mark2old[j]));
				subs&seqleng[j,i]=subs&seqleng[i,j];
				subs&seqleng[i,i]=0;
			%end;
			subs&k[j,i]=subs&k[i,j];
			subs&k[i,i]=0;
        end;
	end;
    %if &print=1 %then print subs&k;;
  %end;
  subs&seqleng=subs%eval(&k-1);
  %if &print=1 %then print subs&seqleng;;
  /*distitp=repeat(0,144,144); /*Matrice temp pour calculer itrativement la distance minimale entre 2 individus*/
  dist=repeat(0,&n,&n);
  do i=2 to &n;
  	do j=1 to (i-1);
		%do k=1 %to &seqleng;
			dist[i,j]=dist[i,j]+subs&k[table[i,&k],table[j,&k]];
		%end;
	end;	
  end;
  return(dist);
  finish;
  use _z;
  %if &weight=0 %then %do;
	read all var {&varlist _wt} into table;
	sortie=subs(table);	
  %end;
  %else %do;
	read all var {&varlist &weight} into table; 
	sortie=subs(table);	
  %end;
  %if &export=0 %then %do;
	create &out from sortie;
 	append from sortie;
  %end;
  %else %do;
	 filename out &outfile;
	 file out ;
	 /*n=nrow(table);*/
	 do i=2 to &n;
		do j=1 to (i-1);
			put @1 (i) 8.0 +1  (j)  8.0 +1 (sortie[i,j]) 20.19 ;
		end;
	 end;
  %end;
quit;
%if &export=0 %then %do;
	data _id;
		length id 3.;
 		do i=1 to &nlist;
   			id=i;
			output;
		end;
		drop i;
	run;
	data &out(type=distance);
		merge _id &out;
	run;
%end;
proc datasets library=work  nolist;
	delete _z;
quit;
options notes;
%put Done!;
%mend;

